@@ -23,13 +23,14 @@
 
 static gboolean
 abort_async_if_port_unusable (MMBaseModem *self,
-                              MMAtSerialPort *port,
+                              MMPortSerialAt *port,
                               GAsyncReadyCallback callback,
                               gpointer user_data)
 {
     GError *error = NULL;
+    gboolean init_sequence_enabled = FALSE;
 
-    /* If no port given, probably the port dissapeared */
+    /* If no port given, probably the port disappeared */
     if (!port) {
         g_simple_async_report_error_in_idle (
             G_OBJECT (self),
@@ -53,8 +54,13 @@ abort_async_if_port_unusable (MMBaseModem *self,
         return FALSE;
     }
 
+    /* Temporarily disable init sequence if we're just sending a
+     * command to a just opened port */
+    g_object_get (port, MM_PORT_SERIAL_AT_INIT_SEQUENCE_ENABLED, &init_sequence_enabled, NULL);
+    g_object_set (port, MM_PORT_SERIAL_AT_INIT_SEQUENCE_ENABLED, FALSE, NULL);
+
     /* Ensure we have a port open during the sequence */
-    if (!mm_serial_port_open (MM_SERIAL_PORT (port), &error)) {
+    if (!mm_port_serial_open (MM_PORT_SERIAL (port), &error)) {
         g_simple_async_report_error_in_idle (
             G_OBJECT (self),
             callback,
@@ -67,6 +73,9 @@ abort_async_if_port_unusable (MMBaseModem *self,
         return FALSE;
     }
 
+    /* Reset previous init sequence state */
+    g_object_set (port, MM_PORT_SERIAL_AT_INIT_SEQUENCE_ENABLED, init_sequence_enabled, NULL);
+
     return TRUE;
 }
 
@@ -81,24 +90,24 @@ modem_cancellable_cancelled (GCancellable *modem_cancellable,
 /* AT sequence handling */
 
 typedef struct {
-    MMBaseModem *self;
-    MMAtSerialPort *port;
-    GCancellable *cancellable;
-    gulong cancelled_id;
-    GCancellable *modem_cancellable;
-    GCancellable *user_cancellable;
+    MMBaseModem                *self;
+    MMPortSerialAt             *port;
+    GCancellable               *cancellable;
+    gulong                      cancelled_id;
+    GCancellable               *modem_cancellable;
+    GCancellable               *user_cancellable;
     const MMBaseModemAtCommand *current;
     const MMBaseModemAtCommand *sequence;
-    GSimpleAsyncResult *simple;
-    gpointer response_processor_context;
-    GDestroyNotify response_processor_context_free;
-    GVariant *result;
+    GSimpleAsyncResult         *simple;
+    gpointer                    response_processor_context;
+    GDestroyNotify              response_processor_context_free;
+    GVariant                   *result;
 } AtSequenceContext;
 
 static void
 at_sequence_context_free (AtSequenceContext *ctx)
 {
-    mm_serial_port_close (MM_SERIAL_PORT (ctx->port));
+    mm_port_serial_close (MM_PORT_SERIAL (ctx->port));
     g_object_unref (ctx->port);
     g_object_unref (ctx->self);
 
@@ -122,15 +131,14 @@ at_sequence_context_free (AtSequenceContext *ctx)
 }
 
 GVariant *
-mm_base_modem_at_sequence_full_finish (MMBaseModem *self,
-                                       GAsyncResult *res,
-                                       gpointer *response_processor_context,
-                                       GError **error)
+mm_base_modem_at_sequence_full_finish (MMBaseModem   *self,
+                                       GAsyncResult  *res,
+                                       gpointer      *response_processor_context,
+                                       GError       **error)
 {
     AtSequenceContext *ctx;
 
-    if (g_simple_async_result_propagate_error (G_SIMPLE_ASYNC_RESULT (res),
-                                               error))
+    if (g_simple_async_result_propagate_error (G_SIMPLE_ASYNC_RESULT (res), error))
         return NULL;
 
     ctx = g_simple_async_result_get_op_res_gpointer (G_SIMPLE_ASYNC_RESULT (res));
@@ -144,79 +152,81 @@ mm_base_modem_at_sequence_full_finish (MMBaseModem *self,
 }
 
 static void
-at_sequence_parse_response (MMAtSerialPort *port,
-                            GString *response,
-                            GError *error,
+at_sequence_parse_response (MMPortSerialAt    *port,
+                            GAsyncResult      *res,
                             AtSequenceContext *ctx)
 {
-    GVariant *result = NULL;
-    GError *result_error = NULL;
-    gboolean continue_sequence;
-    GSimpleAsyncResult *simple;
+    MMBaseModemAtResponseProcessorResult  processor_result;
+    GVariant                             *result = NULL;
+    GError                               *result_error = NULL;
+    GSimpleAsyncResult                   *simple;
+    const gchar                          *response;
+    GError                               *error = NULL;
+
+    response = mm_port_serial_at_command_finish (port, res, &error);
 
     /* Cancelled? */
     if (g_cancellable_is_cancelled (ctx->cancellable)) {
-        g_simple_async_result_set_error (ctx->simple,
-                                         MM_CORE_ERROR,
-                                         MM_CORE_ERROR_CANCELLED,
-                                         "AT sequence was cancelled");
+        g_simple_async_result_set_error (ctx->simple, G_IO_ERROR, G_IO_ERROR_CANCELLED, "AT sequence was cancelled");
+        g_clear_error (&error);
         g_simple_async_result_complete (ctx->simple);
         at_sequence_context_free (ctx);
         return;
     }
 
     if (!ctx->current->response_processor)
-        /* No need to process response, go on to next command */
-        continue_sequence = TRUE;
+        processor_result = MM_BASE_MODEM_AT_RESPONSE_PROCESSOR_RESULT_CONTINUE;
     else {
         const MMBaseModemAtCommand *next = ctx->current + 1;
 
         /* Response processor will tell us if we need to keep on the sequence */
-        continue_sequence = !ctx->current->response_processor (
-            ctx->self,
-            ctx->response_processor_context,
-            ctx->current->command,
-            response ? response->str : NULL,
-            next->command ? FALSE : TRUE,  /* Last command in sequence? */
-            error,
-            &result,
-            &result_error);
-        /* Were we told to abort the sequence? */
-        if (result_error) {
-            g_assert (result == NULL);
-            g_simple_async_result_take_error (ctx->simple, result_error);
-            g_simple_async_result_complete (ctx->simple);
-            at_sequence_context_free (ctx);
-            return;
+        processor_result = ctx->current->response_processor (ctx->self,
+                                                             ctx->response_processor_context,
+                                                             ctx->current->command,
+                                                             response,
+                                                             next->command ? FALSE : TRUE,  /* Last command in sequence? */
+                                                             error,
+                                                             &result,
+                                                             &result_error);
+        switch (processor_result) {
+            case MM_BASE_MODEM_AT_RESPONSE_PROCESSOR_RESULT_CONTINUE:
+                g_assert (!result && !result_error);
+                break;
+            case MM_BASE_MODEM_AT_RESPONSE_PROCESSOR_RESULT_SUCCESS:
+                g_assert (!result_error); /* result is optional */
+                break;
+            case MM_BASE_MODEM_AT_RESPONSE_PROCESSOR_RESULT_FAILURE:
+                /* On failure, complete with error right away */
+                g_assert (!result && result_error); /* result is optional */
+                g_simple_async_result_take_error (ctx->simple, result_error);
+                g_simple_async_result_complete (ctx->simple);
+                at_sequence_context_free (ctx);
+                if (error)
+                    g_error_free (error);
+                return;
+            default:
+                g_assert_not_reached ();
         }
     }
 
-    if (continue_sequence) {
-        g_assert (result == NULL);
+    if (error)
+        g_error_free (error);
+
+    if (processor_result == MM_BASE_MODEM_AT_RESPONSE_PROCESSOR_RESULT_CONTINUE) {
         ctx->current++;
         if (ctx->current->command) {
             /* Schedule the next command in the probing group */
-            if (ctx->current->allow_cached)
-                mm_at_serial_port_queue_command_cached (
-                    ctx->port,
-                    ctx->current->command,
-                    ctx->current->timeout,
-                    FALSE,
-                    ctx->cancellable,
-                    (MMAtSerialResponseFn)at_sequence_parse_response,
-                    ctx);
-            else
-                mm_at_serial_port_queue_command (
-                    ctx->port,
-                    ctx->current->command,
-                    ctx->current->timeout,
-                    FALSE,
-                    ctx->cancellable,
-                    (MMAtSerialResponseFn)at_sequence_parse_response,
-                    ctx);
+            mm_port_serial_at_command (
+                ctx->port,
+                ctx->current->command,
+                ctx->current->timeout,
+                FALSE,
+                ctx->current->allow_cached,
+                ctx->cancellable,
+                (GAsyncReadyCallback)at_sequence_parse_response,
+                ctx);
             return;
         }
-
         /* On last command, end. */
     }
 
@@ -242,14 +252,14 @@ at_sequence_parse_response (MMAtSerialPort *port,
 }
 
 void
-mm_base_modem_at_sequence_full (MMBaseModem *self,
-                                MMAtSerialPort *port,
+mm_base_modem_at_sequence_full (MMBaseModem                *self,
+                                MMPortSerialAt             *port,
                                 const MMBaseModemAtCommand *sequence,
-                                gpointer response_processor_context,
-                                GDestroyNotify response_processor_context_free,
-                                GCancellable *cancellable,
-                                GAsyncReadyCallback callback,
-                                gpointer user_data)
+                                gpointer                    response_processor_context,
+                                GDestroyNotify              response_processor_context_free,
+                                GCancellable               *cancellable,
+                                GAsyncReadyCallback         callback,
+                                gpointer                    user_data)
 {
     AtSequenceContext *ctx;
 
@@ -286,13 +296,14 @@ mm_base_modem_at_sequence_full (MMBaseModem *self,
     }
 
     /* Go on with the first one in the sequence */
-    mm_at_serial_port_queue_command (
+    mm_port_serial_at_command (
         ctx->port,
         ctx->current->command,
         ctx->current->timeout,
         FALSE,
+        ctx->current->allow_cached,
         ctx->cancellable,
-        (MMAtSerialResponseFn)at_sequence_parse_response,
+        (GAsyncReadyCallback)at_sequence_parse_response,
         ctx);
 }
 
@@ -317,7 +328,7 @@ mm_base_modem_at_sequence (MMBaseModem *self,
                            GAsyncReadyCallback callback,
                            gpointer user_data)
 {
-    MMAtSerialPort *port;
+    MMPortSerialAt *port;
     GError *error = NULL;
 
     /* No port given, so we'll try to guess which is best */
@@ -345,76 +356,113 @@ mm_base_modem_at_sequence (MMBaseModem *self,
 /*****************************************************************************/
 /* Response processor helpers */
 
-gboolean
-mm_base_modem_response_processor_string (MMBaseModem *self,
-                                         gpointer none,
-                                         const gchar *command,
-                                         const gchar *response,
-                                         gboolean last_command,
-                                         const GError *error,
-                                         GVariant **result,
-                                         GError **result_error)
+MMBaseModemAtResponseProcessorResult
+mm_base_modem_response_processor_string (MMBaseModem   *self,
+                                         gpointer       none,
+                                         const gchar   *command,
+                                         const gchar   *response,
+                                         gboolean       last_command,
+                                         const GError  *error,
+                                         GVariant     **result,
+                                         GError       **result_error)
 {
     if (error) {
+        *result = NULL;
         *result_error = g_error_copy (error);
-        return FALSE;
+        return MM_BASE_MODEM_AT_RESPONSE_PROCESSOR_RESULT_FAILURE;
     }
 
     *result = g_variant_new_string (response);
-    return TRUE;
+    *result_error = NULL;
+    return MM_BASE_MODEM_AT_RESPONSE_PROCESSOR_RESULT_SUCCESS;
 }
 
-gboolean
-mm_base_modem_response_processor_no_result (MMBaseModem *self,
-                                            gpointer none,
-                                            const gchar *command,
-                                            const gchar *response,
-                                            gboolean last_command,
-                                            const GError *error,
-                                            GVariant **result,
-                                            GError **result_error)
+MMBaseModemAtResponseProcessorResult
+mm_base_modem_response_processor_no_result (MMBaseModem   *self,
+                                            gpointer       none,
+                                            const gchar   *command,
+                                            const gchar   *response,
+                                            gboolean       last_command,
+                                            const GError  *error,
+                                            GVariant     **result,
+                                            GError       **result_error)
 {
     if (error) {
+        *result = NULL;
         *result_error = g_error_copy (error);
-        return FALSE;
+        return MM_BASE_MODEM_AT_RESPONSE_PROCESSOR_RESULT_FAILURE;
     }
 
     *result = NULL;
-    return TRUE;
+    *result_error = NULL;
+    return MM_BASE_MODEM_AT_RESPONSE_PROCESSOR_RESULT_SUCCESS;
 }
 
-gboolean
-mm_base_modem_response_processor_no_result_continue (MMBaseModem *self,
-                                                     gpointer none,
-                                                     const gchar *command,
-                                                     const gchar *response,
-                                                     gboolean last_command,
-                                                     const GError *error,
-                                                     GVariant **result,
-                                                     GError **result_error)
+MMBaseModemAtResponseProcessorResult
+mm_base_modem_response_processor_no_result_continue (MMBaseModem   *self,
+                                                     gpointer       none,
+                                                     const gchar   *command,
+                                                     const gchar   *response,
+                                                     gboolean       last_command,
+                                                     const GError  *error,
+                                                     GVariant     **result,
+                                                     GError       **result_error)
 {
-    if (error)
-        *result_error = g_error_copy (error);
-
-    /* Return FALSE so that we keep on with the next steps in the sequence */
-    return FALSE;
-}
-
-gboolean
-mm_base_modem_response_processor_continue_on_error (MMBaseModem *self,
-                                                    gpointer none,
-                                                    const gchar *command,
-                                                    const gchar *response,
-                                                    gboolean last_command,
-                                                    const GError *error,
-                                                    GVariant **result,
-                                                    GError **result_error)
-{
-    if (error)
-        return FALSE;
-
     *result = NULL;
-    return TRUE;
+
+    if (error) {
+        *result_error = g_error_copy (error);
+        return MM_BASE_MODEM_AT_RESPONSE_PROCESSOR_RESULT_FAILURE;
+    }
+
+    *result_error = NULL;
+    return MM_BASE_MODEM_AT_RESPONSE_PROCESSOR_RESULT_CONTINUE;
+}
+
+MMBaseModemAtResponseProcessorResult
+mm_base_modem_response_processor_continue_on_error (MMBaseModem   *self,
+                                                    gpointer       none,
+                                                    const gchar   *command,
+                                                    const gchar   *response,
+                                                    gboolean       last_command,
+                                                    const GError  *error,
+                                                    GVariant     **result,
+                                                    GError       **result_error)
+{
+    *result = NULL;
+    *result_error = NULL;
+
+    return (error ?
+            MM_BASE_MODEM_AT_RESPONSE_PROCESSOR_RESULT_CONTINUE :
+            MM_BASE_MODEM_AT_RESPONSE_PROCESSOR_RESULT_SUCCESS);
+}
+
+MMBaseModemAtResponseProcessorResult
+mm_base_modem_response_processor_string_ignore_at_errors (MMBaseModem   *self,
+                                                          gpointer       none,
+                                                          const gchar   *command,
+                                                          const gchar   *response,
+                                                          gboolean       last_command,
+                                                          const GError  *error,
+                                                          GVariant     **result,
+                                                          GError       **result_error)
+{
+    if (error) {
+        *result = NULL;
+
+        /* Ignore AT errors (ie, ERROR or CMx ERROR) */
+        if (error->domain != MM_MOBILE_EQUIPMENT_ERROR || last_command) {
+
+            *result_error = g_error_copy (error);
+            return MM_BASE_MODEM_AT_RESPONSE_PROCESSOR_RESULT_FAILURE;
+        }
+
+        *result_error = NULL;
+        return MM_BASE_MODEM_AT_RESPONSE_PROCESSOR_RESULT_CONTINUE;
+    }
+
+    *result = g_variant_new_string (response);
+    return MM_BASE_MODEM_AT_RESPONSE_PROCESSOR_RESULT_SUCCESS;
 }
 
 /*****************************************************************************/
@@ -422,7 +470,7 @@ mm_base_modem_response_processor_continue_on_error (MMBaseModem *self,
 
 typedef struct {
     MMBaseModem *self;
-    MMAtSerialPort *port;
+    MMPortSerialAt *port;
     GCancellable *cancellable;
     gulong cancelled_id;
     GCancellable *modem_cancellable;
@@ -433,7 +481,7 @@ typedef struct {
 static void
 at_command_context_free (AtCommandContext *ctx)
 {
-    mm_serial_port_close (MM_SERIAL_PORT (ctx->port));
+    mm_port_serial_close (MM_PORT_SERIAL (ctx->port));
 
     if (ctx->cancelled_id)
         g_cancellable_disconnect (ctx->modem_cancellable,
@@ -461,39 +509,39 @@ mm_base_modem_at_command_full_finish (MMBaseModem *self,
 }
 
 static void
-at_command_parse_response (MMAtSerialPort *port,
-                           GString *response,
-                           GError *error,
-                           AtCommandContext *ctx)
+at_command_ready (MMPortSerialAt *port,
+                  GAsyncResult *res,
+                  AtCommandContext *ctx)
 {
-    /* Cancelled? */
-    if (g_cancellable_is_cancelled (ctx->cancellable))
-        g_simple_async_result_set_error (ctx->result,
-                                         MM_CORE_ERROR,
-                                         MM_CORE_ERROR_CANCELLED,
-                                         "AT command was cancelled");
+    const gchar *response;
+    GError *error = NULL;
 
+    response = mm_port_serial_at_command_finish (port, res, &error);
+
+    /* Cancelled? */
+    if (g_cancellable_is_cancelled (ctx->cancellable)) {
+        g_simple_async_result_set_error (ctx->result, G_IO_ERROR, G_IO_ERROR_CANCELLED,
+                                         "AT command was cancelled");
+        if (error)
+            g_error_free (error);
+    }
     /* Error coming from the serial port? */
     else if (error)
-        g_simple_async_result_set_from_error (ctx->result, error);
-
+        g_simple_async_result_take_error (ctx->result, error);
     /* Valid string response */
-    else if (response && response->str)
-        g_simple_async_result_set_op_res_gpointer (ctx->result,
-                                                   g_strdup (response->str),
-                                                   g_free);
-
-    /* No response */
+    else if (response)
+        g_simple_async_result_set_op_res_gpointer (ctx->result, (gchar *)response, NULL);
     else
-        g_simple_async_result_set_op_res_gpointer (ctx->result, NULL, NULL);
+        g_assert_not_reached ();
 
+    /* Never in idle! */
     g_simple_async_result_complete (ctx->result);
     at_command_context_free (ctx);
 }
 
 void
 mm_base_modem_at_command_full (MMBaseModem *self,
-                               MMAtSerialPort *port,
+                               MMPortSerialAt *port,
                                const gchar *command,
                                guint timeout,
                                gboolean allow_cached,
@@ -533,24 +581,15 @@ mm_base_modem_at_command_full (MMBaseModem *self,
     }
 
     /* Go on with the command */
-    if (allow_cached)
-        mm_at_serial_port_queue_command_cached (
-            port,
-            command,
-            timeout,
-            is_raw,
-            ctx->cancellable,
-            (MMAtSerialResponseFn)at_command_parse_response,
-            ctx);
-    else
-        mm_at_serial_port_queue_command (
-            port,
-            command,
-            timeout,
-            is_raw,
-            ctx->cancellable,
-            (MMAtSerialResponseFn)at_command_parse_response,
-            ctx);
+    mm_port_serial_at_command (
+        port,
+        command,
+        timeout,
+        is_raw,
+        allow_cached,
+        ctx->cancellable,
+        (GAsyncReadyCallback)at_command_ready,
+        ctx);
 }
 
 const gchar *
@@ -570,7 +609,7 @@ _at_command (MMBaseModem *self,
              GAsyncReadyCallback callback,
              gpointer user_data)
 {
-    MMAtSerialPort *port;
+    MMPortSerialAt *port;
     GError *error = NULL;
 
     /* No port given, so we'll try to guess which is best */
@@ -616,3 +655,9 @@ mm_base_modem_at_command_raw (MMBaseModem *self,
 {
     _at_command (self, command, timeout, allow_cached, TRUE, callback, user_data);
 }
+
+void
+mm_base_modem_at_command_alloc_clear (MMBaseModemAtCommandAlloc *command)
+{
+    g_free (command->command);
+}
